"""
Pokemon name normalization and fuzzing utilities.

This module handles the normalization of Pokemon names to match the expected
data format, including handling special cases, forms, and variations.
"""

from typing import Optional


class PokemonNameNormalizer:
    """
    Handles normalization of Pokemon names for data lookup.
    
    This class contains logic to handle various Pokemon name variations,
    special forms, and edge cases to ensure proper data retrieval.
    """
    
    @staticmethod
    def normalize_name(name: str) -> str:
        """
        Normalize a Pokemon name for data lookup.
        
        Args:
            name: The input Pokemon name
            
        Returns:
            Normalized name that should match the data format
        """
        # Handle spaces and hyphens
        if ' ' in name:
            name = name.strip()
            name = name.replace("   ","-")
            name = name.replace("  ", "-")
            name = name.replace(' ', '-')
            name = name.replace('---', '-')
            name = name.replace('--', '-')
            name = name.replace('- -', '-')
            name = name.replace(' - ', '-')
            name = name.replace(' -', '-')
            name = name.replace('- ', '-')

        

        # Handle special cases
        name = PokemonNameNormalizer._handle_special_cases(name)
        
        # Convert to lowercase at the end
        return name.lower()
    
    @staticmethod
    def _handle_special_cases(name: str) -> str:
        """
        Handle special Pokemon name cases and forms.
        
        Args:
            name: The name to process
            
        Returns:
            Processed name
        """
        name_lower = name.lower()
        
        # Aegislash forms
        if 'aegislash' in name_lower:
            if 'blade' in name_lower:
                return 'aegislash-blade'
            elif 'shield' in name_lower:
                return 'aegislash-shield'
            else:
                return 'aegislash-blade'
        
        if 'alcremie' in name_lower:
            if 'gmax' in name_lower:
                return 'alcremie-gmax'
            else:
                return 'alcremie'
        
        # Arceus forms
        if 'arceus' in name_lower:
            return 'arceus'
        
        # Basculin forms
        if 'basculin' in name_lower:
            if 'blue' in name_lower:
                return 'basculin-blue-striped'
            elif 'red' in name_lower:
                return 'basculin-red-striped'
            elif 'white' in name_lower:
                return 'basculin-white-striped'
            else:
                return 'basculin-blue-striped'
        
        if 'basculegion' in name_lower:
            if 'f' in name_lower:
                return 'basculegion-female'
            elif 'm' in name_lower:
                return 'basculegion-male'
            else:
                return 'basculegion-female'
        
        # Calyrex forms
        if 'calyrex' in name_lower:
            if 'ice' in name_lower:
                return 'calyrex-ice'
            elif 'shadow' in name_lower:
                return 'calyrex-shadow'
            else:
                return 'calyrex-ice'
        
        # Castform forms
        if 'castform' in name_lower:
            if 'rainy' in name_lower:
                return 'castform-rainy'
            elif 'snowy' in name_lower:
                return 'castform-snowy'
            elif 'sunny' in name_lower:
                return 'castform-sunny'
            else:
                return 'castform-sunny'
        
        if 'cramorant' in name_lower:
            if 'gulping' in name_lower:
                return 'cramorant-gulping'
            elif 'gorging' in name_lower:
                return 'cramorant-gorging'
            else:
                return 'cramorant'
        
        # Darmanitan forms
        if 'darmanitan' in name_lower:
            if 'zen' in name_lower:
                return 'darmanitan-zen'
            else:
                return 'darmanitan-standard'
        
        # Deoxys forms
        if 'deoxys' in name_lower:
            if 'attack' in name_lower:
                return 'deoxys-attack'
            elif 'defense' in name_lower:
                return 'deoxys-defense'
            elif 'speed' in name_lower:
                return 'deoxys-speed'
            elif 'normal' in name_lower:
                return 'deoxys-normal'
            else:
                return 'deoxys'
        
        # Dialga forms
        if 'dialga' in name_lower:
            if 'origin' in name_lower:
                return 'dialga-origin'
            else:
                return 'dialga'
        
        # Dudunsparce forms
        if 'dudunsparce' in name_lower:
            if 'three' in name_lower:
                return 'dudunsparce-three-segment'
            else:
                return 'dudunsparce-two-segment'
        
        # Eevee forms
        if 'eevee' in name_lower:
            if 'starter' in name_lower:
                return 'eevee-starter'
            else:
                return 'eevee'
        
        # Eiscue forms
        if 'eiscue' in name_lower:
            if 'noice' in name_lower:
                return 'eiscue-noice'
            else:
                return 'eiscue-ice'
        
        # Enamorus forms
        if 'enamorus' in name_lower:
            if 'therian' in name_lower:
                return 'enamorus-therian'
            else:
                return 'enamorus-incarnate'
        
        # Floette forms
        if 'floette' in name_lower:
            if 'eternal' in name_lower:
                return 'floette-eternal'
            else:
                return 'floette'
        
        # Florges forms
        if 'florges' in name_lower:
            return 'florges'
        
        if 'gastrodon' in name_lower:
            return 'gastrodon'
        
        # Gimmighoul forms
        if 'gimmighoul' in name_lower:
            if 'roaming' in name_lower:
                return 'gimmighoul-roaming'
            else:
                return 'gimmighoul'
        
        # Giratina forms
        if 'giratina' in name_lower:
            if 'origin' in name_lower:
                return 'giratina-origin'
            else:
                return 'giratina-altered'
        
        # Gourgeist forms
        if 'gourgeist' in name_lower:
            if 'large' in name_lower:
                return 'gourgeist-large'
            elif 'small' in name_lower:
                return 'gourgeist-small'
            elif 'super' in name_lower:
                return 'gourgeist-super'
            else:
                return 'gourgeist-average'
        
        # Greninja forms
        if 'greninja' in name_lower:
            if 'ash' in name_lower:
                return 'greninja-ash'
            elif 'battle' in name_lower:
                return 'greninja-battle-bond'
            else:
                return 'greninja'
        
        # Hoopa forms
        if 'hoopa' in name_lower:
            if 'unbound' in name_lower:
                return 'hoopa-unbound'
            else:
                return 'hoopa'
        
        # Indeedee forms
        if 'indeedee' in name_lower:
            if 'f' in name_lower:
                return 'indeedee-female'
            elif 'm' in name_lower:
                return 'indeedee-male'
            else:
                return 'indeedee-female'
        
        # Keldeo forms
        if 'keldeo' in name_lower:
            if 'resolute' in name_lower:
                return 'keldeo-resolute'
            else:
                return 'keldeo-ordinary'
        
        # Koraidon forms
        if 'koraidon' in name_lower:
            if 'gliding' in name_lower:
                return 'koraidon-gliding-build'
            elif 'limited' in name_lower:
                return 'koraidon-limited-build'
            elif 'sprinting' in name_lower:
                return 'koraidon-sprinting-build'
            elif 'swimming' in name_lower:
                return 'koraidon-swimming-build'
            else:
                return 'koraidon'
        
        # Kyurem forms
        if 'kyurem' in name_lower:
            if 'black' in name_lower:
                return 'kyurem-black'
            elif 'white' in name_lower:
                return 'kyurem-white'
            else:
                return 'kyurem'
        
        # Landorus forms
        if 'landorus' in name_lower:
            if 'therian' in name_lower:
                return 'landorus-therian'
            else:
                return 'landorus-incarnate'
        
        # Lycanroc forms
        if 'lycanroc' in name_lower:
            if 'dusk' in name_lower:
                return 'lycanroc-dusk'
            elif 'midday' in name_lower:
                return 'lycanroc-midday'
            else:
                return 'lycanroc-midnight'
        
        # Magearna forms
        if 'magearna' in name_lower:
            if 'original' in name_lower:
                return 'magearna-original'
            else:
                return 'magearna'
        
        # Maushold forms
        if 'maushold' in name_lower:
            return 'maushold'
        
        if 'meowstic' in name_lower:
            if 'f' in name_lower:
                return 'meowstic-female'
            else:
                return 'meowstic-male'
        
        if 'meloetta' in name_lower:
            if 'aria' in name_lower:
                return 'meloetta-aria'
            elif 'pirouette' in name_lower:
                return 'meloetta-pirouette'
            else:
                return 'meloetta-aria'
        
        # Mimikyu forms
        if 'mimikyu' in name_lower:
            if 'busted' in name_lower:
                return 'mimikyu-busted'
            else:
                return 'mimikyu-disguised'
        
        # Minior forms
        if 'minior' in name_lower:
            if 'blue' in name_lower:
                if 'meteor' in name_lower:
                    return 'minior-blue-meteor'
                else:
                    return 'minior-blue'
            elif 'green' in name_lower:
                if 'meteor' in name_lower:
                    return 'minior-green-meteor'
                else:
                    return 'minior-green'
            elif 'indigo' in name_lower:
                if 'meteor' in name_lower:
                    return 'minior-indigo-meteor'
                else:
                    return 'minior-indigo'
            elif 'orange' in name_lower:
                if 'meteor' in name_lower:
                    return 'minior-orange-meteor'
                else:
                    return 'minior-orange'
            elif 'red' in name_lower:
                if 'meteor' in name_lower:
                    return 'minior-red-meteor'
                else:
                    return 'minior-red'
            elif 'violet' in name_lower:
                if 'meteor' in name_lower:
                    return 'minior-violet-meteor'
                else:
                    return 'minior-violet'
            elif 'yellow' in name_lower:
                if 'meteor' in name_lower:
                    return 'minior-yellow-meteor'
                else:
                    return 'minior-yellow'
            else:
                return 'minior-red'
        
        # Miraidon forms
        if 'miraidon' in name_lower:
            if 'aquatic' in name_lower:
                return 'miraidon-aquatic-mode'
            elif 'drive' in name_lower:
                return 'miraidon-drive-mode'
            elif 'glide' in name_lower:
                return 'miraidon-glide-mode'
            elif 'low' in name_lower:
                return 'miraidon-low-power-mode'
            else:
                return 'miraidon'
        
        if 'morpeko' in name_lower:
            if 'full-belly' in name_lower:
                return 'morpeko-full-belly'
            else:
                return 'morpeko-hangry'
        
        if 'necrozma' in name_lower:
            if 'dawn' in name_lower:
                return 'necrozma-dawn'
            elif 'dusk' in name_lower:
                return 'necrozma-dusk'
            elif 'ultra' in name_lower:
                return 'necrozma-ultra'
            else:
                return 'necrozma'
        
        # Nidoran forms
        if 'nidoran' in name_lower:
            if 'f' in name_lower:
                return 'nidoran-f'
            else:
                return 'nidoran-m'
        
        # Ogerpon forms
        if 'ogerpon' in name_lower:
            if 'cornerstone' in name_lower:
                return 'ogerpon-cornerstone-mask'
            elif 'wellspring' in name_lower:
                return 'ogerpon-wellspring-mask'
            elif 'hearthflame' in name_lower:
                return 'ogerpon-hearthflame-mask'
            else:
                return 'ogerpon'
        
        if 'oinkologne' in name_lower:
            if 'f' in name_lower:
                return 'oinkologne-female'
            elif 'm' in name_lower:
                return 'oinkologne-male'
            else:
                return 'oinkologne-female'
        
        #Oricorio forms
        if 'oricorio' in name_lower:
            if 'baile' in name_lower:
                return 'oricorio-baile'
            elif 'pa' in name_lower:
                return 'oricorio-pau'
            elif 'sensu' in name_lower:
                return 'oricorio-sensu'
            elif 'pom' in name_lower:
                return 'oricorio-pom-pom'
            else:
                return 'oricorio-baile'
        
        # Palafin forms
        if 'palafin' in name_lower:
            if 'hero' in name_lower:
                return 'palafin-hero'
            else:
                return 'palafin-zero'
        
        # Palkia forms
        if 'palkia' in name_lower:
            if 'origin' in name_lower:
                return 'palkia-origin'
            else:
                return 'palkia'
        
        # Pikachu forms
        if 'pikachu' in name_lower:
            if 'alola' in name_lower:
                return 'pikachu-alola-cap'
            elif 'belle' in name_lower:
                return 'pikachu-belle'
            elif 'cosplay' in name_lower:
                return 'pikachu-cosplay'
            elif 'hoenn' in name_lower:
                return 'pikachu-hoenn-cap'
            elif 'kalos' in name_lower:
                return 'pikachu-kalos-cap'
            elif 'libre' in name_lower:
                return 'pikachu-libre'
            elif 'original' in name_lower:
                return 'pikachu-original-cap'
            elif 'partner' in name_lower:
                return 'pikachu-partner-cap'
            elif 'phd' in name_lower:
                return 'pikachu-phd'
            elif 'pop' in name_lower:
                return 'pikachu-pop-star'
            elif 'rock' in name_lower:
                return 'pikachu-rock-star'
            elif 'sinnoh' in name_lower:
                return 'pikachu-sinnoh-cap'
            elif 'starter' in name_lower:
                return 'pikachu-starter'
            elif 'unova' in name_lower:
                return 'pikachu-unova-cap'
            elif 'world' in name_lower:
                return 'pikachu-world-cap'
            else:
                return 'pikachu'
        
        if 'polteageist' in name_lower:
            return 'polteageist'
        
        if 'poltchageist' in name_lower:
            return 'poltchageist'
        
        if 'porygon-z' in name_lower:
            return 'porygon-z'
        
        # Pumpkaboo forms
        if 'pumpkaboo' in name_lower:
            if 'large' in name_lower:
                return 'pumpkaboo-large'
            elif 'small' in name_lower:
                return 'pumpkaboo-small'
            elif 'super' in name_lower:
                return 'pumpkaboo-super'
            else:
                return 'pumpkaboo-average'
        
        # Rockruff forms
        if 'rockruff' in name_lower:
            if 'own' in name_lower:
                return 'rockruff-own-tempo'
            else:
                return 'rockruff'
        
        # Rotom forms
        if 'rotom' in name_lower:
            if 'fan' in name_lower:
                return 'rotom-fan'
            elif 'frost' in name_lower:
                return 'rotom-frost'
            elif 'heat' in name_lower:
                return 'rotom-heat'
            elif 'mow' in name_lower:
                return 'rotom-mow'
            elif 'wash' in name_lower:
                return 'rotom-wash'
            else:
                return 'rotom'
        
        if 'sawsbuck' in name_lower:
            return 'sawsbuck'
        
        # Shaymin forms
        if 'shaymin' in name_lower:
            if 'sky' in name_lower:
                return 'shaymin-sky'
            else:
                return 'shaymin-land'
        
        if 'sinistcha' in name_lower:
            return 'sinistcha'
        
        if 'sinistea' in name_lower:
            return 'sinistea'
        
        # Squawkabilly forms
        if 'squawkabilly' in name_lower:
            if 'blue' in name_lower:
                return 'squawkabilly-blue-plumage'
            elif 'red' in name_lower:
                return 'squawkabilly-red-plumage'
            elif 'white' in name_lower:
                return 'squawkabilly-white-plumage'
            elif 'green' in name_lower:
                return 'squawkabilly-green-plumage'
            else:
                return 'squawkabilly-white-plumage'
        
        # Tapu forms
        if 'tapu' in name_lower:
            if 'bulu' in name_lower:
                return 'tapu-bulu'
            elif 'fini' in name_lower:
                return 'tapu-fini'
            elif 'koko' in name_lower:
                return 'tapu-koko'
            elif 'lele' in name_lower:
                return 'tapu-lele'
            else:
                return 'tapu-koko'
        
        # Tauros Paldea forms
        if 'tauros' in name_lower and 'paldea' in name_lower:
            if 'blaze' in name_lower:
                return 'tauros-paldea-blaze-breed'
            elif 'aqua' in name_lower:
                return 'tauros-paldea-aqua-breed'
            elif 'combat' in name_lower:
                return 'tauros-paldea-combat-breed'
            else:
                return 'tauros-paldea-blaze-breed'
        
        # Tatsugiri forms
        if 'tatsugiri' in name_lower:
            if 'curly' in name_lower:
                return 'tatsugiri-curly'
            elif 'droopy' in name_lower:
                return 'tatsugiri-droopy'
            else:
                return 'tatsugiri-stretchy'
        
        # Terapagos forms
        if 'terapagos' in name_lower:
            if 'stellar' in name_lower:
                return 'terapagos-stellar'
            elif 'terastal' in name_lower:
                return 'terapagos-terastal'
            else:
                return 'terapagos'
        
        # Thundurus forms
        if 'thundurus' in name_lower:
            if 'therian' in name_lower:
                return 'thundurus-therian'
            else:
                return 'thundurus-incarnate'
        
        # Tornadus forms
        if 'tornadus' in name_lower:
            if 'therian' in name_lower:
                return 'tornadus-therian'
            else:
                return 'tornadus-incarnate'
        
        # Toxtricity forms
        if 'toxtricity' in name_lower:
            if 'amped' in name_lower:
                return 'toxtricity-amped'
            else:
                return 'toxtricity-low-key'
        
        if 'urshifu' in name_lower:
            if 'single' in name_lower:
                return 'urshifu-single-strike'
            elif 'rapid' in name_lower:
                return 'urshifu-rapid-strike'
            else:
                return 'urshifu-single-strike'
        
        # Ursaluna forms
        if 'ursaluna' in name_lower:
            if 'bloodmoon' in name_lower:
                return 'ursaluna-bloodmoon'
            else:
                return 'ursaluna'
        
        if 'vivillon' in name_lower:
            return 'vivillon'
        
        # Wishiwashi forms
        if 'wishiwashi' in name_lower:
            if 'school' in name_lower:
                return 'wishiwashi-school'
            else:
                return 'wishiwashi-solo'
        
        # Wormadam forms
        if 'wormadam' in name_lower:
            if 'sandy' in name_lower:
                return 'wormadam-sandy'
            elif 'trash' in name_lower:
                return 'wormadam-trash'
            else:
                return 'wormadam-plant'
        
        if 'zacian' in name_lower:
            if 'crowned' in name_lower:
                return 'zacian-crowned'
            else:
                return 'zacian'
        
        if 'zamazenta' in name_lower:
            if 'crowned' in name_lower:
                return 'zamazenta-crowned'
            else:
                return 'zamazenta'
        
        # Zarude forms
        if 'zarude' in name_lower:
            if 'dada' in name_lower:
                return 'zarude-dada'
            else:
                return 'zarude'
        
        # Zygarde forms
        if 'zygarde' in name_lower:
            if 'complete' in name_lower:
                return 'zygarde-complete'
            elif '10' in name_lower:
                if 'power' in name_lower:
                    return 'zygarde-10-power-construct'
                else:
                    return 'zygarde-10'
            elif '50' in name_lower:
                if 'power' in name_lower:
                    return 'zygarde-50-power-construct'
                else:
                    return 'zygarde-50'
            else:
                return 'zygarde-50'
        
        return name 


"""
Comprehensive test suite for LocalDex core functionality.

This module tests the main LocalDex class including initialization,
data retrieval, caching, search capabilities, and error handling.
"""

import pytest
import tempfile
import json
import os
from unittest.mock import Mock, patch, MagicMock
from pathlib import Path

from localdex.core import LocalDex
from localdex.models import Pokemon, Move, Ability, Item, BaseStats
from localdex.exceptions import (
    PokemonNotFoundError, MoveNotFoundError, AbilityNotFoundError,
    ItemNotFoundError, SearchError
)


class TestLocalDexInitialization:
    """Test LocalDex initialization and configuration."""
    
    def test_init_with_defaults(self):
        """Test LocalDex initialization with default parameters."""
        dex = LocalDex()
        assert dex.enable_caching is True
        assert dex.data_loader is not None
        assert len(dex._pokemon_cache) == 0
        assert len(dex._move_cache) == 0
        assert len(dex._ability_cache) == 0
        assert len(dex._item_cache) == 0
    
    def test_init_with_custom_data_path(self):
        """Test LocalDex initialization with custom data path."""
        with tempfile.TemporaryDirectory() as temp_dir:
            dex = LocalDex(data_path=temp_dir)
            assert dex.data_dir == temp_dir
    
    def test_init_with_data_dir_alias(self):
        """Test LocalDex initialization with data_dir parameter (backward compatibility)."""
        with tempfile.TemporaryDirectory() as temp_dir:
            dex = LocalDex(data_dir=temp_dir)
            assert dex.data_dir == temp_dir
    
    def test_init_with_caching_disabled(self):
        """Test LocalDex initialization with caching disabled."""
        dex = LocalDex(enable_caching=False)
        assert dex.enable_caching is False
        # Should not build indexes when caching is disabled
        assert len(dex._pokemon_by_type) == 0
        assert len(dex._pokemon_by_generation) == 0
    
    def test_init_builds_indexes_when_caching_enabled(self):
        """Test that indexes are built when caching is enabled."""
        # Mock the data loader to return some test data
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            
            # Mock get_all_pokemon to return test data
            test_pokemon = [
                Pokemon(
                    id=1, name="Bulbasaur", types=["Grass", "Poison"],
                    base_stats=BaseStats(45, 49, 49, 65, 65, 45),
                    generation=1
                ),
                Pokemon(
                    id=4, name="Charmander", types=["Fire"],
                    base_stats=BaseStats(39, 52, 43, 60, 50, 65),
                    generation=1
                )
            ]
            
            with patch.object(LocalDex, 'get_all_pokemon', return_value=test_pokemon):
                with patch.object(LocalDex, 'get_all_moves', return_value=[]):
                    dex = LocalDex(enable_caching=True)
                    
                    # Check that indexes were built
                    assert "grass" in dex._pokemon_by_type
                    assert "poison" in dex._pokemon_by_type
                    assert "fire" in dex._pokemon_by_type
                    assert 1 in dex._pokemon_by_generation
                    assert "bulbasaur" in dex._pokemon_by_type["grass"]
                    assert "charmander" in dex._pokemon_by_type["fire"]


class TestLocalDexPokemonRetrieval:
    """Test Pokemon retrieval functionality."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)  # Disable caching for cleaner tests
            return dex
    
    def test_get_pokemon_by_id(self, mock_dex):
        """Test getting Pokemon by ID."""
        # Mock data
        pokemon_data = {
            "id": 25,
            "name": "Pikachu",
            "types": ["Electric"],
            "baseStats": {"hp": 35, "attack": 55, "defense": 40, 
                         "special_attack": 50, "special_defense": 50, "speed": 90},
            "generation": 1
        }
        
        mock_dex.data_loader.load_pokemon_by_id.return_value = pokemon_data
        
        pokemon = mock_dex.get_pokemon_by_id(25)
        
        assert pokemon.id == 25
        assert pokemon.name == "Pikachu"
        assert pokemon.types == ["Electric"]
        assert pokemon.base_stats.hp == 35
        assert pokemon.base_stats.speed == 90
        assert pokemon.generation == 1
        
        mock_dex.data_loader.load_pokemon_by_id.assert_called_once_with(25)
    
    def test_get_pokemon_by_id_not_found(self, mock_dex):
        """Test getting Pokemon by ID when not found."""
        mock_dex.data_loader.load_pokemon_by_id.return_value = None
        
        with pytest.raises(PokemonNotFoundError) as exc_info:
            mock_dex.get_pokemon_by_id(99999)
        
        assert "99999" in str(exc_info.value)
    
    def test_get_pokemon_by_name(self, mock_dex):
        """Test getting Pokemon by name."""
        pokemon_data = {
            "id": 6,
            "name": "Charizard",
            "types": ["Fire", "Flying"],
            "baseStats": {"hp": 78, "attack": 84, "defense": 78, 
                         "special_attack": 109, "special_defense": 85, "speed": 100},
            "generation": 1
        }
        
        mock_dex.data_loader.load_pokemon_by_name.return_value = pokemon_data
        
        pokemon = mock_dex.get_pokemon_by_name("Charizard")
        
        assert pokemon.id == 6
        assert pokemon.name == "Charizard"
        assert pokemon.types == ["Fire", "Flying"]
        assert pokemon.base_stats.special_attack == 109
        
        mock_dex.data_loader.load_pokemon_by_name.assert_called_once_with("Charizard")
    
    def test_get_pokemon_by_name_case_insensitive(self, mock_dex):
        """Test getting Pokemon by name with different case."""
        pokemon_data = {
            "id": 1,
            "name": "Bulbasaur",
            "types": ["Grass", "Poison"],
            "baseStats": {"hp": 45, "attack": 49, "defense": 49, 
                         "special_attack": 65, "special_defense": 65, "speed": 45}
        }
        
        mock_dex.data_loader.load_pokemon_by_name.return_value = pokemon_data
        
        pokemon = mock_dex.get_pokemon_by_name("BULBASAUR")
        
        assert pokemon.name == "Bulbasaur"
        mock_dex.data_loader.load_pokemon_by_name.assert_called_with("BULBASAUR")
    
    def test_get_pokemon_by_name_not_found(self, mock_dex):
        """Test getting Pokemon by name when not found."""
        mock_dex.data_loader.load_pokemon_by_name.return_value = None
        
        with pytest.raises(PokemonNotFoundError) as exc_info:
            mock_dex.get_pokemon_by_name("NonexistentPokemon")
        
        assert "NonexistentPokemon" in str(exc_info.value)
    
    def test_get_pokemon_with_id(self, mock_dex):
        """Test get_pokemon method with ID."""
        pokemon_data = {
            "id": 150,
            "name": "Mewtwo",
            "types": ["Psychic"],
            "baseStats": {"hp": 106, "attack": 110, "defense": 90, 
                         "special_attack": 154, "special_defense": 90, "speed": 130}
        }
        
        mock_dex.data_loader.load_pokemon_by_id.return_value = pokemon_data
        
        pokemon = mock_dex.get_pokemon(150)
        
        assert pokemon.id == 150
        assert pokemon.name == "Mewtwo"
        mock_dex.data_loader.load_pokemon_by_id.assert_called_once_with(150)
    
    def test_get_pokemon_with_name(self, mock_dex):
        """Test get_pokemon method with name."""
        pokemon_data = {
            "id": 151,
            "name": "Mew",
            "types": ["Psychic"],
            "baseStats": {"hp": 100, "attack": 100, "defense": 100, 
                         "special_attack": 100, "special_defense": 100, "speed": 100}
        }
        
        mock_dex.data_loader.load_pokemon_by_name.return_value = pokemon_data
        
        pokemon = mock_dex.get_pokemon("Mew")
        
        assert pokemon.id == 151
        assert pokemon.name == "Mew"
        mock_dex.data_loader.load_pokemon_by_name.assert_called_once_with("Mew")


class TestLocalDexMoveRetrieval:
    """Test move retrieval functionality."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)
            return dex
    
    def test_get_move(self, mock_dex):
        """Test getting a move by name."""
        move_data = {
            "name": "Thunderbolt",
            "type": "Electric",
            "category": "Special",
            "basePower": 90,
            "accuracy": 100,
            "pp": 15,
            "priority": 0,
            "target": "normal",
            "desc": "A strong electric blast"
        }
        
        mock_dex.data_loader.load_move.return_value = move_data
        
        move = mock_dex.get_move("Thunderbolt")
        
        assert move.name == "Thunderbolt"
        assert move.type == "Electric"
        assert move.category == "Special"
        assert move.base_power == 90
        assert move.accuracy == 100
        assert move.pp == 15
        
        mock_dex.data_loader.load_move.assert_called_once_with("Thunderbolt")
    
    def test_get_move_not_found(self, mock_dex):
        """Test getting a move that doesn't exist."""
        mock_dex.data_loader.load_move.return_value = None
        
        with pytest.raises(MoveNotFoundError) as exc_info:
            mock_dex.get_move("NonexistentMove")
        
        assert "NonexistentMove" in str(exc_info.value)


class TestLocalDexAbilityRetrieval:
    """Test ability retrieval functionality."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)
            return dex
    
    def test_get_ability(self, mock_dex):
        """Test getting an ability by name."""
        ability_data = {
            "name": "Levitate",
            "description": "This Pokemon is immune to Ground-type moves.",
            "generation": 3,
            "rating": 3.5
        }
        
        mock_dex.data_loader.load_ability.return_value = ability_data
        
        ability = mock_dex.get_ability("Levitate")
        
        assert ability.name == "Levitate"
        assert ability.description == "This Pokemon is immune to Ground-type moves."
        assert ability.generation == 3
        assert ability.rating == 3.5
        
        mock_dex.data_loader.load_ability.assert_called_once_with("Levitate")
    
    def test_get_ability_not_found(self, mock_dex):
        """Test getting an ability that doesn't exist."""
        mock_dex.data_loader.load_ability.return_value = None
        
        with pytest.raises(AbilityNotFoundError) as exc_info:
            mock_dex.get_ability("NonexistentAbility")
        
        assert "NonexistentAbility" in str(exc_info.value)


class TestLocalDexItemRetrieval:
    """Test item retrieval functionality."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)
            return dex
    
    def test_get_item(self, mock_dex):
        """Test getting an item by name."""
        item_data = {
            "name": "Leftovers",
            "description": "Restores HP each turn.",
            "generation": 2,
            "category": "held"
        }
        
        mock_dex.data_loader.load_item.return_value = item_data
        
        item = mock_dex.get_item("Leftovers")
        
        assert item.name == "Leftovers"
        assert item.description == "Restores HP each turn."
        assert item.generation == 2
        assert item.category == "held"
        
        mock_dex.data_loader.load_item.assert_called_once_with("Leftovers")
    
    def test_get_item_not_found(self, mock_dex):
        """Test getting an item that doesn't exist."""
        mock_dex.data_loader.load_item.return_value = None
        
        with pytest.raises(ItemNotFoundError) as exc_info:
            mock_dex.get_item("NonexistentItem")
        
        assert "NonexistentItem" in str(exc_info.value)


class TestLocalDexCaching:
    """Test caching functionality."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=True)
            return dex
    
    def test_pokemon_caching(self, mock_dex):
        """Test that Pokemon are cached after retrieval."""
        pokemon_data = {
            "id": 1,
            "name": "Bulbasaur",
            "types": ["Grass", "Poison"],
            "baseStats": {"hp": 45, "attack": 49, "defense": 49, 
                         "special_attack": 65, "special_defense": 65, "speed": 45}
        }
        
        mock_dex.data_loader.load_pokemon_by_name.return_value = pokemon_data
        
        # First call should load from data
        pokemon1 = mock_dex.get_pokemon_by_name("Bulbasaur")
        
        # Second call should use cache
        pokemon2 = mock_dex.get_pokemon_by_name("Bulbasaur")
        
        assert pokemon1 is pokemon2  # Same object from cache
        mock_dex.data_loader.load_pokemon_by_name.assert_called_once()
    
    def test_move_caching(self, mock_dex):
        """Test that moves are cached after retrieval."""
        move_data = {
            "name": "Tackle",
            "type": "Normal",
            "category": "Physical",
            "basePower": 40,
            "accuracy": 100,
            "pp": 35
        }
        
        mock_dex.data_loader.load_move.return_value = move_data
        
        # First call should load from data
        move1 = mock_dex.get_move("Tackle")
        
        # Second call should use cache
        move2 = mock_dex.get_move("Tackle")
        
        assert move1 is move2  # Same object from cache
        mock_dex.data_loader.load_move.assert_called_once()
    
    def test_clear_cache(self, mock_dex):
        """Test clearing all caches."""
        # Add some data to caches
        mock_dex._pokemon_cache["test"] = Mock()
        mock_dex._move_cache["test"] = Mock()
        mock_dex._ability_cache["test"] = Mock()
        mock_dex._item_cache["test"] = Mock()
        
        mock_dex.clear_cache()
        
        assert len(mock_dex._pokemon_cache) == 0
        assert len(mock_dex._move_cache) == 0
        assert len(mock_dex._ability_cache) == 0
        assert len(mock_dex._item_cache) == 0
    
    def test_get_cache_stats(self, mock_dex):
        """Test getting cache statistics."""
        # Add some data to caches
        mock_dex._pokemon_cache["pokemon1"] = Mock()
        mock_dex._pokemon_cache["pokemon2"] = Mock()
        mock_dex._move_cache["move1"] = Mock()
        mock_dex._pokemon_id_cache["1"] = Mock()
        
        stats = mock_dex.get_cache_stats()
        
        assert stats["pokemon_by_name"] == 2
        assert stats["pokemon_by_id"] == 1
        assert stats["moves"] == 1
        assert stats["abilities"] == 0
        assert stats["items"] == 0


class TestLocalDexSearch:
    """Test search functionality."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with test data."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            
            dex = LocalDex(enable_caching=True)
            
            # Add test Pokemon to indexes
            dex._pokemon_by_type = {
                "fire": {"charizard", "charmander", "charmeleon"},
                "water": {"squirtle", "wartortle", "blastoise"},
                "grass": {"bulbasaur", "ivysaur", "venusaur"}
            }
            
            dex._pokemon_by_generation = {
                1: {"bulbasaur", "charmander", "squirtle", "charizard", "venusaur", "blastoise"}
            }
            
            return dex
    
    def test_search_pokemon_by_type(self, mock_dex):
        """Test searching Pokemon by type."""
        # Mock get_pokemon_by_name to return test Pokemon
        def mock_get_pokemon(name):
            pokemon = Mock()
            pokemon.name = name
            pokemon.types = ["Fire"] if "char" in name else ["Water"] if "squirt" in name else ["Grass"]
            pokemon.base_stats = Mock()
            pokemon.base_stats.attack = 80
            pokemon.base_stats.special_attack = 85
            pokemon.base_stats.speed = 100
            pokemon.base_stats.hp = 78
            pokemon.generation = 1
            pokemon.is_legendary = False
            pokemon.is_mythical = False
            return pokemon
        
        mock_dex.get_pokemon_by_name = mock_get_pokemon
        
        results = mock_dex.search_pokemon(type="Fire")
        
        assert len(results) == 3
        names = {pokemon.name for pokemon in results}
        assert names == {"charizard", "charmander", "charmeleon"}
    
    def test_search_pokemon_by_generation(self, mock_dex):
        """Test searching Pokemon by generation."""
        def mock_get_pokemon(name):
            pokemon = Mock()
            pokemon.name = name
            pokemon.types = ["Normal"]
            pokemon.base_stats = Mock()
            pokemon.base_stats.attack = 80
            pokemon.base_stats.special_attack = 85
            pokemon.base_stats.speed = 100
            pokemon.base_stats.hp = 78
            pokemon.generation = 1
            pokemon.is_legendary = False
            pokemon.is_mythical = False
            return pokemon
        
        mock_dex.get_pokemon_by_name = mock_get_pokemon
        
        results = mock_dex.search_pokemon(generation=1)
        
        assert len(results) == 6
        names = {pokemon.name for pokemon in results}
        assert "bulbasaur" in names
        assert "charmander" in names
        assert "squirtle" in names
    
    def test_search_pokemon_by_stats(self, mock_dex):
        """Test searching Pokemon by stat filters."""
        def mock_get_pokemon(name):
            pokemon = Mock()
            pokemon.name = name
            pokemon.types = ["Normal"]
            pokemon.base_stats = Mock()
            pokemon.base_stats.attack = 100 if "strong" in name else 50
            pokemon.base_stats.special_attack = 90 if "special" in name else 60
            pokemon.base_stats.speed = 120 if "fast" in name else 70
            pokemon.base_stats.hp = 80
            pokemon.generation = 1
            pokemon.is_legendary = False
            pokemon.is_mythical = False
            return pokemon

        # Mock the data loader to return proper data
        mock_dex.data_loader.load_all_pokemon.return_value = []
        mock_dex.get_pokemon_by_name = mock_get_pokemon
        mock_dex._pokemon_by_type["normal"] = {"strongpokemon", "specialpokemon", "fastpokemon"}

        results = mock_dex.search_pokemon(type="Normal", min_attack=90)

        assert len(results) == 1
        assert results[0].name == "strongpokemon"
    
    def test_search_pokemon_multiple_filters(self, mock_dex):
        """Test searching Pokemon with multiple filters."""
        def mock_get_pokemon(name):
            pokemon = Mock()
            pokemon.name = name
            pokemon.types = ["Fire"] if "char" in name else ["Water"]
            pokemon.base_stats = Mock()
            pokemon.base_stats.attack = 100 if "charizard" in name else 50
            pokemon.base_stats.special_attack = 85
            pokemon.base_stats.speed = 100
            pokemon.base_stats.hp = 78
            pokemon.generation = 1
            pokemon.is_legendary = False
            pokemon.is_mythical = False
            return pokemon
        
        mock_dex.get_pokemon_by_name = mock_get_pokemon
        
        results = mock_dex.search_pokemon(type="Fire", min_attack=90)
        
        assert len(results) == 1
        assert results[0].name == "charizard"
    
    def test_search_pokemon_name_contains(self, mock_dex):
        """Test searching Pokemon by partial name match."""
        def mock_get_pokemon(name):
            pokemon = Mock()
            pokemon.name = name
            pokemon.types = ["Normal"]
            pokemon.base_stats = Mock()
            pokemon.base_stats.attack = 80
            pokemon.base_stats.special_attack = 85
            pokemon.base_stats.speed = 100
            pokemon.base_stats.hp = 78
            pokemon.generation = 1
            pokemon.is_legendary = False
            pokemon.is_mythical = False
            return pokemon

        # Mock the data loader to return proper data
        mock_dex.data_loader.load_all_pokemon.return_value = []
        mock_dex.get_pokemon_by_name = mock_get_pokemon
        mock_dex._pokemon_by_type["normal"] = {"charizard", "charmander", "charmeleon"}

        results = mock_dex.search_pokemon(type="Normal", name_contains="char")

        assert len(results) == 3
        names = {pokemon.name for pokemon in results}
        assert names == {"charizard", "charmander", "charmeleon"}
    
    def test_search_pokemon_legendary_filter(self, mock_dex):
        """Test searching Pokemon by legendary status."""
        def mock_get_pokemon(name):
            pokemon = Mock()
            pokemon.name = name
            pokemon.types = ["Psychic"]
            pokemon.base_stats = Mock()
            pokemon.base_stats.attack = 110
            pokemon.base_stats.special_attack = 154
            pokemon.base_stats.speed = 130
            pokemon.base_stats.hp = 106
            pokemon.generation = 1
            pokemon.is_legendary = True if "mewtwo" in name else False
            pokemon.is_mythical = False
            return pokemon

        # Mock the data loader to return proper data
        mock_dex.data_loader.load_all_pokemon.return_value = []
        mock_dex.get_pokemon_by_name = mock_get_pokemon
        mock_dex._pokemon_by_type["psychic"] = {"mewtwo", "alakazam"}

        results = mock_dex.search_pokemon(type="Psychic", is_legendary=True)

        assert len(results) == 1
        assert results[0].name == "mewtwo"


class TestLocalDexDataCreation:
    """Test data object creation from raw data."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)
            return dex
    
    def test_create_pokemon_from_data(self, mock_dex):
        """Test creating Pokemon object from raw data."""
        pokemon_data = {
            "id": 25,
            "name": "Pikachu",
            "types": ["Electric"],
            "baseStats": {
                "hp": 35,
                "attack": 55,
                "defense": 40,
                "special_attack": 50,
                "special_defense": 50,
                "speed": 90
            },
            "height": 0.4,
            "weight": 6.0,
            "color": "Yellow",
            "abilities": {"0": "Static", "1": "Lightning Rod"},
            "generation": 1,
            "description": "A mouse Pokemon",
            "isLegendary": False,
            "isMythical": False,
            "isUltraBeast": False
        }
        
        pokemon = mock_dex._create_pokemon_from_data(pokemon_data)
        
        assert pokemon.id == 25
        assert pokemon.name == "Pikachu"
        assert pokemon.types == ["Electric"]
        assert pokemon.base_stats.hp == 35
        assert pokemon.base_stats.speed == 90
        assert pokemon.height == 0.4
        assert pokemon.weight == 6.0
        assert pokemon.color == "Yellow"
        assert pokemon.abilities == {"0": "Static", "1": "Lightning Rod"}
        assert pokemon.generation == 1
        assert pokemon.description == "A mouse Pokemon"
        assert pokemon.is_legendary is False
        assert pokemon.is_mythical is False
        assert pokemon.is_ultra_beast is False
    
    def test_create_move_from_data(self, mock_dex):
        """Test creating Move object from raw data."""
        move_data = {
            "name": "Thunderbolt",
            "type": "Electric",
            "category": "Special",
            "basePower": 90,
            "accuracy": 100,
            "pp": 15,
            "priority": 0,
            "target": "normal",
            "desc": "A strong electric blast",
            "shortDesc": "Strong electric attack",
            "contestType": "Cool",
            "critRatio": 1,
            "flags": {"protect": True, "mirror": True},
            "drain": 0,
            "isZ": False,
            "num": 85
        }
        
        move = mock_dex._create_move_from_data(move_data)
        
        assert move.name == "Thunderbolt"
        assert move.type == "Electric"
        assert move.category == "Special"
        assert move.base_power == 90
        assert move.accuracy == 100
        assert move.pp == 15
        assert move.priority == 0
        assert move.target == "normal"
        assert move.description == "A strong electric blast"
        assert move.short_description == "Strong electric attack"
        assert move.contest_type == "Cool"
        assert move.crit_ratio == 1
        assert move.flags == {"protect": True, "mirror": True}
        assert move.drain == 0
        assert move.z_move is False
        assert move.generation == 85
    
    def test_create_ability_from_data(self, mock_dex):
        """Test creating Ability object from raw data."""
        ability_data = {
            "name": "Levitate",
            "description": "This Pokemon is immune to Ground-type moves.",
            "short_description": "Immune to Ground moves",
            "generation": 3,
            "rating": 3.5,
            "num": 26,
            "effect": "This Pokemon is immune to Ground-type moves."
        }
        
        ability = mock_dex._create_ability_from_data(ability_data)
        
        assert ability.name == "Levitate"
        assert ability.description == "This Pokemon is immune to Ground-type moves."
        assert ability.short_description == "Immune to Ground moves"
        assert ability.generation == 3
        assert ability.rating == 3.5
        assert ability.num == 26
        assert ability.effect == "This Pokemon is immune to Ground-type moves."
    
    def test_create_item_from_data(self, mock_dex):
        """Test creating Item object from raw data."""
        item_data = {
            "name": "Leftovers",
            "description": "Restores HP each turn.",
            "shortDesc": "Restores HP",
            "gen": 2,
            "num": 234,
            "spritenum": 2,
            "fling": {"basePower": 10},
            "megaStone": False,
            "megaEvolves": None,
            "zMove": False,
            "zMoveType": None,
            "zMoveFrom": None,
            "itemUser": None,
            "onPlate": False,
            "onDrive": False,
            "onMemory": False,
            "forcedForme": None,
            "isNonstandard": False,
            "category": "held"
        }
        
        item = mock_dex._create_item_from_data(item_data)
        
        assert item.name == "Leftovers"
        assert item.description == "Restores HP each turn."
        assert item.short_description == "Restores HP"
        assert item.generation == 2
        assert item.num == 234
        assert item.spritenum == 2
        assert item.fling == {"basePower": 10}
        assert item.mega_stone is False
        assert item.mega_evolves is None
        assert item.z_move is False
        assert item.category == "held"


class TestLocalDexBulkOperations:
    """Test bulk data retrieval operations."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)
            return dex
    
    def test_get_all_pokemon(self, mock_dex):
        """Test getting all Pokemon."""
        pokemon_data_list = [
            {"id": 1, "name": "Bulbasaur", "types": ["Grass", "Poison"], 
             "baseStats": {"hp": 45, "attack": 49, "defense": 49, 
                          "special_attack": 65, "special_defense": 65, "speed": 45}},
            {"id": 4, "name": "Charmander", "types": ["Fire"], 
             "baseStats": {"hp": 39, "attack": 52, "defense": 43, 
                          "special_attack": 60, "special_defense": 50, "speed": 65}}
        ]
        
        mock_dex.data_loader.load_all_pokemon.return_value = pokemon_data_list
        
        pokemon_list = mock_dex.get_all_pokemon()
        
        assert len(pokemon_list) == 2
        assert pokemon_list[0].name == "Bulbasaur"
        assert pokemon_list[1].name == "Charmander"
        
        mock_dex.data_loader.load_all_pokemon.assert_called_once()
    
    def test_get_all_moves(self, mock_dex):
        """Test getting all moves."""
        move_data_list = [
            {"name": "Tackle", "type": "Normal", "category": "Physical", 
             "basePower": 40, "accuracy": 100, "pp": 35},
            {"name": "Thunderbolt", "type": "Electric", "category": "Special", 
             "basePower": 90, "accuracy": 100, "pp": 15}
        ]
        
        mock_dex.data_loader.load_all_moves.return_value = move_data_list
        
        move_list = mock_dex.get_all_moves()
        
        assert len(move_list) == 2
        assert move_list[0].name == "Tackle"
        assert move_list[1].name == "Thunderbolt"
        
        mock_dex.data_loader.load_all_moves.assert_called_once()
    
    def test_get_all_abilities(self, mock_dex):
        """Test getting all abilities."""
        ability_data_list = [
            {"name": "Levitate", "description": "Immune to Ground moves"},
            {"name": "Static", "description": "May paralyze on contact"}
        ]
        
        mock_dex.data_loader.load_all_abilities.return_value = ability_data_list
        
        ability_list = mock_dex.get_all_abilities()
        
        assert len(ability_list) == 2
        assert ability_list[0].name == "Levitate"
        assert ability_list[1].name == "Static"
        
        mock_dex.data_loader.load_all_abilities.assert_called_once()
    
    def test_get_all_items(self, mock_dex):
        """Test getting all items."""
        item_data_list = [
            {"name": "Leftovers", "description": "Restores HP each turn"},
            {"name": "Choice Band", "description": "Boosts Attack but locks moves"}
        ]
        
        mock_dex.data_loader.load_all_items.return_value = item_data_list
        
        item_list = mock_dex.get_all_items()
        
        assert len(item_list) == 2
        assert item_list[0].name == "Leftovers"
        assert item_list[1].name == "Choice Band"
        
        mock_dex.data_loader.load_all_items.assert_called_once()


class TestLocalDexStatGuesser:
    """Test stat guessing functionality."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked dependencies."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)
            return dex
    
    def test_create_stat_guesser(self, mock_dex):
        """Test creating a BattleStatGuesser instance."""
        with patch('localdex.core.BattleStatGuesser') as mock_guesser_class:
            mock_guesser = Mock()
            mock_guesser_class.return_value = mock_guesser
            
            guesser = mock_dex.create_stat_guesser("gen9ou")
            
            mock_guesser_class.assert_called_once_with("gen9ou", mock_dex)
            assert guesser == mock_guesser
    
    def test_guess_pokemon_stats(self, mock_dex):
        """Test guessing Pokemon stats."""
        with patch('localdex.core.BattleStatGuesser') as mock_guesser_class:
            mock_guesser = Mock()
            mock_guesser_class.return_value = mock_guesser
            
            pokemon_set = Mock()
            expected_result = {"role": "attacker", "evs": {"atk": 252, "spe": 252}}
            mock_guesser.guess.return_value = expected_result
            
            result = mock_dex.guess_pokemon_stats(pokemon_set, "gen9ou")
            
            assert result == expected_result
            mock_guesser.guess.assert_called_once_with(pokemon_set)
    
    def test_optimize_pokemon_stats(self, mock_dex):
        """Test optimizing Pokemon stats."""
        with patch('localdex.core.battle_stat_optimizer') as mock_optimizer:
            pokemon_set = Mock()
            expected_result = {"optimized_evs": {"atk": 252, "spe": 252}}
            mock_optimizer.return_value = expected_result
            
            result = mock_dex.optimize_pokemon_stats(pokemon_set, "gen9ou")
            
            assert result == expected_result
            mock_optimizer.assert_called_once_with(pokemon_set, "gen9ou", mock_dex)


class TestLocalDexErrorHandling:
    """Test error handling and edge cases."""
    
    @pytest.fixture
    def mock_dex(self):
        """Create a LocalDex instance with mocked data loader."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            dex = LocalDex(enable_caching=False)
            return dex
    
    def test_search_error_handling(self, mock_dex):
        """Test that search errors are properly handled."""
        # Mock get_all_pokemon to raise an exception
        mock_dex.get_all_pokemon = Mock(side_effect=Exception("Test error"))
        
        with pytest.raises(SearchError) as exc_info:
            mock_dex.search_pokemon(type="Fire")
        
        assert "Error during Pokemon search" in str(exc_info.value)
    
    def test_index_building_error_handling(self, mock_dex):
        """Test that index building errors don't crash initialization."""
        # Mock get_all_pokemon to raise an exception during index building
        with patch.object(LocalDex, 'get_all_pokemon', side_effect=Exception("Index error")):
            # Should not raise an exception, just continue without indexes
            dex = LocalDex(enable_caching=True)
            assert len(dex._pokemon_by_type) == 0
            assert len(dex._pokemon_by_generation) == 0
    
    def test_missing_data_handling(self, mock_dex):
        """Test handling of missing or incomplete data."""
        # Test with minimal Pokemon data
        minimal_data = {
            "id": 1,
            "name": "TestPokemon",
            "types": ["Normal"],
            "baseStats": {"hp": 50, "attack": 50, "defense": 50, 
                         "special_attack": 50, "special_defense": 50, "speed": 50}
        }
        
        pokemon = mock_dex._create_pokemon_from_data(minimal_data)
        
        assert pokemon.id == 1
        assert pokemon.name == "TestPokemon"
        assert pokemon.types == ["Normal"]
        # Optional fields should have default values
        assert pokemon.height is None
        assert pokemon.weight is None
        assert pokemon.generation is None
        assert pokemon.is_legendary is False
        assert pokemon.is_mythical is False
        assert pokemon.is_ultra_beast is False


class TestLocalDexIntegration:
    """Integration tests for LocalDex functionality."""
    
    def test_full_workflow(self):
        """Test a complete workflow with LocalDex."""
        with patch('localdex.core.DataLoader') as mock_loader_class:
            mock_loader = Mock()
            mock_loader_class.return_value = mock_loader
            
            # Mock data
            pokemon_data = {
                "id": 25,
                "name": "Pikachu",
                "types": ["Electric"],
                "baseStats": {"hp": 35, "attack": 55, "defense": 40, 
                             "special_attack": 50, "special_defense": 50, "speed": 90},
                "generation": 1
            }
            
            move_data = {
                "name": "Thunderbolt",
                "type": "Electric",
                "category": "Special",
                "basePower": 90,
                "accuracy": 100,
                "pp": 15
            }
            
            mock_loader.load_pokemon_by_name.return_value = pokemon_data
            mock_loader.load_move.return_value = move_data
            
            # Create LocalDex instance
            dex = LocalDex(enable_caching=True)
            
            # Get Pokemon and move
            pokemon = dex.get_pokemon("Pikachu")
            move = dex.get_move("Thunderbolt")
            
            # Verify data
            assert pokemon.name == "Pikachu"
            assert pokemon.types == ["Electric"]
            assert move.name == "Thunderbolt"
            assert move.type == "Electric"
            
            # Check cache stats
            stats = dex.get_cache_stats()
            assert stats["pokemon_by_name"] == 1
            assert stats["moves"] == 1
            
            # Clear cache
            dex.clear_cache()
            stats = dex.get_cache_stats()
            assert stats["pokemon_by_name"] == 0
            assert stats["moves"] == 0


if __name__ == "__main__":
    pytest.main([__file__])


"""
Comprehensive test suite for Pokemon name normalization functionality.

This module tests the PokemonNameNormalizer class to ensure it properly handles
various Pokemon name variations including capitalization, spaces, hyphens,
gender indicators, forms, and special cases. Tests verify that normalized names
actually correspond to valid Pokemon in the data.
"""

import pytest
from localdex.name_normalizer import PokemonNameNormalizer
from localdex.core import LocalDex


class TestNameNormalizerIntegration:
    """Test name normalization with actual LocalDex data validation."""
    
    @pytest.fixture
    def dex(self):
        """Create a LocalDex instance for validation."""
        return LocalDex()
    
    @pytest.fixture
    def normalizer(self):
        """Create a PokemonNameNormalizer instance."""
        return PokemonNameNormalizer()
    
    def validate_normalized_name(self, dex, normalizer, input_name):
        """Helper method to validate that a normalized name corresponds to a real Pokemon."""
        normalized = normalizer.normalize_name(input_name)
        
        # Try to get the Pokemon by the normalized name
        try:
            pokemon = dex.get_pokemon_by_name(normalized)
            return pokemon is not None, normalized, pokemon.name if pokemon else None
        except Exception:
            return False, normalized, None
    
    def test_basic_name_normalization(self, dex, normalizer):
        """Test basic name normalization with validation."""
        test_cases = [
            "bulbasaur",
            "CHARIZARD", 
            "MeWtWo",
            "PikAChU",
            "Blastoise",
            "squirtle",
            "pikachu",
            "mewtwo"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_spaces_to_hyphens(self, dex, normalizer):
        """Test that spaces are converted to hyphens and result in valid Pokemon."""
        test_cases = [
            "mr mime",
            "mime jr", 
            "type null",
            "porygon z",
            "ho oh"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_gender_variations(self, dex, normalizer):
        """Test gender-specific Pokemon name variations."""
        test_cases = [
            "nidoran",
            "nidoran m",
            "nidoran-m",
            "nidoran male",
            "nidoran f",
            "nidoran-f",
            "nidoran female",
            "indeedee",
            "indeedee f",
            "indeedee-f",
            "indeedee female",
            "indeedee m",
            "indeedee-m",
            "indeedee male",
            "meowstic",
            "meowstic m",
            "meowstic-m",
            "meowstic male",
            "meowstic f",
            "meowstic-f",
            "meowstic female",
            "basculegion",
            "basculegion f",
            "basculegion-f",
            "basculegion female",
            "basculegion m",
            "basculegion-m",
            "basculegion male",
            "oinkologne",
            "oinkologne f",
            "oinkologne-f",
            "oinkologne female",
            "oinkologne m",
            "oinkologne-m",
            "oinkologne male"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_form_variations(self, dex, normalizer):
        """Test Pokemon form variations."""
        test_cases = [
            "aegislash",
            "aegislash blade",
            "aegislash-blade",
            "aegislash shield",
            "aegislash-shield",
            "basculin",
            "basculin blue",
            "basculin-blue",
            "basculin red",
            "basculin-red",
            "basculin white",
            "basculin-white",
            "deoxys",
            "deoxys normal",
            "deoxys-normal",
            "deoxys attack",
            "deoxys-attack",
            "deoxys defense",
            "deoxys-defense",
            "deoxys speed",
            "deoxys-speed",
            "castform",
            "castform sunny",
            "castform-sunny",
            "castform rainy",
            "castform-rainy",
            "castform snowy",
            "castform-snowy",
            "rotom",
            "rotom fan",
            "rotom-fan",
            "rotom frost",
            "rotom-frost",
            "rotom heat",
            "rotom-heat",
            "rotom mow",
            "rotom-mow",
            "rotom wash",
            "rotom-wash",
            "lycanroc",
            "lycanroc midnight",
            "lycanroc-midnight",
            "lycanroc midday",
            "lycanroc-midday",
            "lycanroc dusk",
            "lycanroc-dusk"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_legendary_forms(self, dex, normalizer):
        """Test legendary Pokemon form variations."""
        test_cases = [
            "kyurem",
            "kyurem black",
            "kyurem-black",
            "kyurem white",
            "kyurem-white",
            "giratina",
            "giratina altered",
            "giratina-altered",
            "giratina origin",
            "giratina-origin",
            "dialga",
            "dialga origin",
            "dialga-origin",
            "palkia",
            "palkia origin",
            "palkia-origin"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_size_forms(self, dex, normalizer):
        """Test size-based form variations."""
        test_cases = [
            "pumpkaboo",
            "pumpkaboo average",
            "pumpkaboo-average",
            "pumpkaboo small",
            "pumpkaboo-small",
            "pumpkaboo large",
            "pumpkaboo-large",
            "pumpkaboo super",
            "pumpkaboo-super",
            "gourgeist",
            "gourgeist average",
            "gourgeist-average",
            "gourgeist small",
            "gourgeist-small",
            "gourgeist large",
            "gourgeist-large",
            "gourgeist super",
            "gourgeist-super"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_mixed_case_with_spaces_and_hyphens(self, dex, normalizer):
        """Test names with mixed case, spaces, and hyphens."""
        test_cases = [
            "Mr Mime",
            "Mime Jr",
            "Type Null",
            "Porygon Z",
            "Ho Oh",
            "Deoxys Attack",
            "Kyurem Black",
            "Giratina Origin"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_extra_whitespace_handling(self, dex, normalizer):
        """Test handling of extra whitespace."""
        test_cases = [
            "  bulbasaur  ",
            "  mr   mime  ",
            "  deoxys   attack  ",
            "\tkyurem\tblack\t",
            "\n  giratina  origin  \n"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_numbers_in_names(self, dex, normalizer):
        """Test Pokemon names with numbers."""
        test_cases = [
            "porygon-z",
            "porygon z",
            "zygarde-10",
            "zygarde 10",
            "zygarde-50",
            "zygarde 50",
            "zygarde-complete",
            "zygarde complete"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_regional_variants(self, dex, normalizer):
        """Test regional variant Pokemon."""
        test_cases = [
            "raichu alola",
            "raichu-alola",
            "vulpix alola",
            "vulpix-alola",
            "ninetales alola",
            "ninetales-alola",
            "diglett alola",
            "diglett-alola",
            "meowth alola",
            "meowth-alola",
            "meowth galar",
            "meowth-galar",
            "ponyta galar",
            "ponyta-galar",
            "rapidash galar",
            "rapidash-galar",
            "farfetchd galar",
            "farfetchd-galar",
            "corsola galar",
            "corsola-galar",
            "growlithe hisui",
            "growlithe-hisui",
            "arcanine hisui",
            "arcanine-hisui",
            "voltorb hisui",
            "voltorb-hisui",
            "electrode hisui",
            "electrode-hisui"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_mega_evolutions(self, dex, normalizer):
        """Test Mega Evolution variations."""
        test_cases = [
            "venusaur mega",
            "venusaur-mega",
            "charizard mega x",
            "charizard-mega-x",
            "charizard mega y",
            "charizard-mega-y",
            "blastoise mega",
            "blastoise-mega",
            "alakazam mega",
            "alakazam-mega",
            "gengar mega",
            "gengar-mega"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_gigantamax_forms(self, dex, normalizer):
        """Test Gigantamax form variations."""
        test_cases = [
            "butterfree gmax",
            "butterfree-gmax",
            "charizard gmax",
            "charizard-gmax",
            "blastoise gmax",
            "blastoise-gmax",
            "alcremie gmax",
            "alcremie-gmax"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_common_pokemon_names(self, dex, normalizer):
        """Test normalization with common Pokemon names."""
        common_names = [
            "bulbasaur", "ivysaur", "venusaur", "charmander", "charmeleon",
            "charizard", "squirtle", "wartortle", "blastoise", "caterpie",
            "metapod", "butterfree", "weedle", "kakuna", "beedrill",
            "pidgey", "pidgeotto", "pidgeot", "rattata", "raticate",
            "spearow", "fearow", "ekans", "arbok", "pikachu", "raichu"
        ]
        
        for name in common_names:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, name)
            assert is_valid, f"'{name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_legendary_pokemon(self, dex, normalizer):
        """Test normalization with legendary Pokemon."""
        legendary_names = [
            "articuno", "zapdos", "moltres", "mewtwo", "mew",
            "raikou", "entei", "suicune", "lugia", "ho-oh",
            "celebi", "regirock", "regice", "registeel", "latias",
            "latios", "kyogre", "groudon", "rayquaza", "jirachi",
            "deoxys", "dialga", "palkia", "giratina", "arceus"
        ]
        
        for name in legendary_names:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, name)
            assert is_valid, f"'{name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_edge_cases(self, dex, normalizer):
        """Test edge cases and error conditions."""
        # Test empty string
        normalized = normalizer.normalize_name("")
        assert normalized == ""
        
        # Test whitespace-only string
        normalized = normalizer.normalize_name("   ")
        assert normalized == ""
        
        # Test single character
        normalized = normalizer.normalize_name("a")
        assert normalized == "a"
        
        # Test very long name
        long_name = "a" * 1000
        result = normalizer.normalize_name(long_name)
        assert result == long_name.lower()
    
    def test_unicode_characters(self, dex, normalizer):
        """Test Pokemon names with Unicode characters."""
        test_cases = [
            "jangmo-o",
            "hakamo-o", 
            "kommo-o"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_complex_form_variations(self, dex, normalizer):
        """Test complex form variations based on actual filenames."""
        test_cases = [
            "mimikyu", "mimikyu busted", "mimikyu-busted", "mimikyu disguised", "mimikyu-disguised",
            "minior", "minior red", "minior-red", "minior blue", "minior-blue",
            "squawkabilly", "squawkabilly white plumage", "squawkabilly-white-plumage",
            "tatsugiri", "tatsugiri curly", "tatsugiri-curly", "tatsugiri droopy", "tatsugiri-droopy",
            "dudunsparce", "dudunsparce three segment", "dudunsparce-three-segment",
            "maushold", "maushold family of three", "maushold-family-of-three",
            "palafin", "palafin zero", "palafin-zero", "palafin hero", "palafin-hero",
            "miraidon", "miraidon aquatic mode", "miraidon-aquatic-mode",
            "koraidon", "koraidon gliding build", "koraidon-gliding-build",
            "ogerpon", "ogerpon cornerstone mask", "ogerpon-cornerstone-mask",
            "terapagos", "terapagos stellar", "terapagos-stellar",
            "ursaluna", "ursaluna bloodmoon", "ursaluna-bloodmoon",
            "urshifu", "urshifu single strike", "urshifu-single-strike",
            "toxtricity", "toxtricity amped", "toxtricity-amped",
            "wormadam", "wormadam plant", "wormadam-plant",
            "thundurus", "thundurus incarnate", "thundurus-incarnate",
            "zygarde", "zygarde 10", "zygarde-10", "zygarde complete", "zygarde-complete",
            "shaymin", "shaymin land", "shaymin-land", "shaymin sky", "shaymin-sky",
            "oricorio", "oricorio baile", "oricorio-baile", "oricorio pom pom", "oricorio-pom-pom",
            "tauros", "tauros paldea blaze breed", "tauros-paldea-blaze-breed",
            "gimmighoul", "gimmighoul roaming", "gimmighoul-roaming"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_hisui_regional_variants(self, dex, normalizer):
        """Test Hisui regional variant Pokemon."""
        test_cases = [
            "typhlosion hisui", "typhlosion-hisui", "decidueye hisui", "decidueye-hisui",
            "samurott hisui", "samurott-hisui", "braviary hisui", "braviary-hisui",
            "lilligant hisui", "lilligant-hisui", "sliggoo hisui", "sliggoo-hisui",
            "goodra hisui", "goodra-hisui", "avalugg hisui", "avalugg-hisui",
            "qwilfish hisui", "qwilfish-hisui", "sneasel hisui", "sneasel-hisui",
            "zoroark hisui", "zoroark-hisui", "zorua hisui", "zorua-hisui",
            "overqwil", "sneasler", "kleavor", "ursaluna", "ursaluna-bloodmoon",
            "wyrdeer", "basculegion", "basculegion-female", "basculegion-male"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_alolan_regional_variants(self, dex, normalizer):
        """Test Alolan regional variant Pokemon."""
        test_cases = [
            "raichu alola", "raichu-alola", "vulpix alola", "vulpix-alola",
            "ninetales alola", "ninetales-alola", "diglett alola", "diglett-alola",
            "meowth alola", "meowth-alola", "persian alola", "persian-alola",
            "geodude alola", "geodude-alola", "grimer alola", "grimer-alola",
            "exeggutor alola", "exeggutor-alola", "marowak alola", "marowak-alola",
            "raticate alola", "raticate-alola", "rattata alola", "rattata-alola",
            "sandshrew alola", "sandshrew-alola", "sandslash alola", "sandslash-alola"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_galarian_regional_variants(self, dex, normalizer):
        """Test Galarian regional variant Pokemon."""
        test_cases = [
            "meowth galar", "meowth-galar", "ponyta galar", "ponyta-galar",
            "rapidash galar", "rapidash-galar", "farfetchd galar", "farfetchd-galar",
            "corsola galar", "corsola-galar", "weezing galar", "weezing-galar",
            "mr mime galar", "mr-mime-galar", "mr rime", "mr-rime",
            "slowpoke galar", "slowpoke-galar", "slowbro galar", "slowbro-galar",
            "slowking galar", "slowking-galar", "articuno galar", "articuno-galar",
            "zapdos galar", "zapdos-galar", "moltres galar", "moltres-galar",
            "zigzagoon galar", "zigzagoon-galar", "linoone galar", "linoone-galar"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_mega_evolution_variations(self, dex, normalizer):
        """Test Mega Evolution variations."""
        test_cases = [
            "venusaur mega", "venusaur-mega", "venusaur gmax", "venusaur-gmax",
            "charizard mega x", "charizard-mega-x", "charizard mega y", "charizard-mega-y", "charizard gmax", "charizard-gmax",
            "blastoise mega", "blastoise-mega", "blastoise gmax", "blastoise-gmax",
            "alakazam mega", "alakazam-mega", "gengar mega", "gengar-mega", "gengar gmax", "gengar-gmax",
            "mewtwo mega x", "mewtwo-mega-x", "mewtwo mega y", "mewtwo-mega-y",
            "rayquaza mega", "rayquaza-mega"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_gigantamax_variations(self, dex, normalizer):
        """Test Gigantamax variations."""
        test_cases = [
            "butterfree gmax", "butterfree-gmax", "charizard gmax", "charizard-gmax",
            "blastoise gmax", "blastoise-gmax", "alcremie gmax", "alcremie-gmax",
            "appletun gmax", "appletun-gmax", "centiskorch gmax", "centiskorch-gmax",
            "coalossal gmax", "coalossal-gmax", "copperajah gmax", "copperajah-gmax",
            "duraludon gmax", "duraludon-gmax", "eevee gmax", "eevee-gmax",
            "flapple gmax", "flapple-gmax", "garbodor gmax", "garbodor-gmax",
            "gengar gmax", "gengar-gmax", "grimmsnarl gmax", "grimmsnarl-gmax",
            "hatterene gmax", "hatterene-gmax", "inteleon gmax", "inteleon-gmax",
            "kingler gmax", "kingler-gmax", "lapras gmax", "lapras-gmax",
            "machamp gmax", "machamp-gmax", "meowth gmax", "meowth-gmax",
            "melmetal gmax", "melmetal-gmax", "orbeetle gmax", "orbeetle-gmax",
            "sandaconda gmax", "sandaconda-gmax", "snorlax gmax", "snorlax-gmax",
            "toxtricity amped gmax", "toxtricity-amped-gmax", "toxtricity low key gmax", "toxtricity-low-key-gmax",
            "urshifu single strike gmax", "urshifu-single-strike-gmax", "urshifu rapid strike gmax", "urshifu-rapid-strike-gmax",
            "venusaur gmax", "venusaur-gmax"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_iron_pokemon(self, dex, normalizer):
        """Test Iron Pokemon (Paradox forms)."""
        test_cases = [
            "iron boulder", "iron-boulder", "iron bundle", "iron-bundle",
            "iron crown", "iron-crown", "iron hands", "iron-hands",
            "iron jugulis", "iron-jugulis", "iron leaves", "iron-leaves",
            "iron moth", "iron-moth", "iron thorns", "iron-thorns",
            "iron treads", "iron-treads", "iron valiant", "iron-valiant"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_paradox_pokemon(self, dex, normalizer):
        """Test Paradox Pokemon."""
        test_cases = [
            "great tusk", "great-tusk", "scream tail", "scream-tail",
            "brute bonnet", "brute-bonnet", "flutter mane", "flutter-mane",
            "slither wing", "slither-wing", "sandy shocks", "sandy-shocks",
            "roaring moon", "roaring-moon", "iron valiant", "iron-valiant",
            "walking wake", "walking-wake", "raging bolt", "raging-bolt",
            "gouging fire", "gouging-fire"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")
    
    def test_special_event_forms(self, dex, normalizer):
        """Test special event form Pokemon."""
        test_cases = [
            "pikachu", "pikachu starter", "pikachu-starter", "pikachu gmax", "pikachu-gmax",
            "pikachu alola cap", "pikachu-alola-cap", "pikachu hoenn cap", "pikachu-hoenn-cap",
            "pikachu sinnoh cap", "pikachu-sinnoh-cap", "pikachu unova cap", "pikachu-unova-cap",
            "pikachu kalos cap", "pikachu-kalos-cap", "pikachu original cap", "pikachu-original-cap",
            "pikachu partner cap", "pikachu-partner-cap", "pikachu world cap", "pikachu-world-cap",
            "pikachu cosplay", "pikachu-cosplay", "pikachu belle", "pikachu-belle",
            "pikachu pop star", "pikachu-pop-star", "pikachu rock star", "pikachu-rock-star",
            "pikachu phd", "pikachu-phd", "pikachu libre", "pikachu-libre",
            "eevee starter", "eevee-starter", "eevee gmax", "eevee-gmax",
            "floette eternal", "floette-eternal", "rockruff own tempo", "rockruff-own-tempo"
        ]
        
        for input_name in test_cases:
            is_valid, normalized, actual_name = self.validate_normalized_name(dex, normalizer, input_name)
            assert is_valid, f"'{input_name}' normalized to '{normalized}' but no Pokemon found"
            print(f"✓ '{input_name}' -> '{normalized}' -> '{actual_name}'")


if __name__ == "__main__":
    pytest.main([__file__]) 